/* approximate subset sum file packer
   by eadmaster
   
   released under the terms of the latest GPL license.
*/
   
   
//#define _FILE_OFFSET_BITS 64
//#define _LARGE_FILES
//#define _LARGEFILE_SOURCE
//#define _LARGEFILE64_SOURCE
// now the uintmax_t type should be 64-bit long even on 32-bit systems

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <dirent.h>
#include <inttypes.h>
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
	// define a fnmatch alternative
	// http://stackoverflow.com/questions/2364302/fnmatch-for-windows-in-c
	#include <windows.h>
	#include <shlwapi.h>
	#define fnmatch(x, y, z)  ( PathMatchSpec(y, x) ? 0 : 1 )
#else
	#include <fnmatch.h>
#endif



const uintmax_t GIGABYTE=1024*1024*1024;
const uintmax_t MEGABYTE=1024*1024;
const uintmax_t KILOBYTE=1024;
// TODO: fixare overflow in "_to_bytes", sostituire uintmax_t<->off_t ?


const uintmax_t FD_SIZE=1457664;
const uintmax_t CDR_SIZE=736962560;
//const uintmax_t CD_R_74_MIN=737280000;
//const uintmax_t CD_R_80_MIN=840960000;	
const uintmax_t DVD_PLUS_R_SIZE=4700372992ll; // DVD+R
const uintmax_t DVD_PLUS_R_DL_SIZE=8547991552ll; // DVD+R DL
const uintmax_t DVD_MINUS_R_SIZE=4707319808ll; // DVD-R
const uintmax_t DVD_MINUS_R_DL_SIZE=8543666176ll; // DVD-R DL


/*
inline int max(int a, int b)
{
	if(a>=b) return a;
	else return b;
}*/


// TODO: usare gnulib "human.h" (usato anche da gnuwin32 "du") | altra lib

uintmax_t _to_bytes(char* s)
/*
	convert a human readable byte string (with suffixes "G, M, K")
	into RAW bytes count
	
	also supports special constants like fd, cd-r, dvd-r, etc.
*/
{
	long double W;
	char unit='b';

	// convert "s" to lower case for case-insensitive string comparison
	int i;
	for(i = 0; s[ i ]; i++)
		s[i] = tolower(s[ i ]);

	if(sscanf( s, "%Lf%c", &W, &unit )<=0) // TODO: buffer overflow checks "%3Lf%1c%*s"
	{
		// a special constant is used
		if(strcmp(s,"fd")==0)
			return FD_SIZE;
		else if(strcmp(s,"cdr")==0)
			return CDR_SIZE;
		else if(strcmp(s,"cd-r")==0)
			return CDR_SIZE;
		else if(strcmp(s,"dvd+r")==0)
			return DVD_PLUS_R_SIZE;
		else if(strcmp(s,"dvd+rdl")==0)
			return DVD_PLUS_R_DL_SIZE;
		else if(strcmp(s,"dvd-r")==0)
			return DVD_MINUS_R_SIZE;
		else if(strcmp(s,"dvd-rdl")==0)
			return DVD_MINUS_R_DL_SIZE;
		else
			return(-1);		
	} else {
		// explicit size is used
		switch(toupper(unit))
		{
			/*
			case 'T': // terabytes
				W*=1024;
				// no break, fall down
			*/
			
			case 'G': // gigabytes
				W*=1024;
				// no break, fall down
			
			case 'M': // megabytes
				W*=1024;
				// no break, fall down
			
			case 'K': // kilobytes
				W*=1024;
			
			case 'B':
				// do nothing
				break;
			
			default:
				// invalid unit specified
				return(-1);
		}
printf("%Lf\n", W); // 2FIX: overflow?
		return W;
	}

	return(-1);
}


char* _to_human(uintmax_t n)
/* convert "n" RAW bytes count
   into a human-readable string
   (allocated internally)
*/
{
	char* r = calloc(1000, sizeof(char)); // TODO: replace "1000"
		
	if(n>=GIGABYTE) {
		sprintf(r, "%.3gGB", (n/(double)GIGABYTE));
	}
	else if(n>=MEGABYTE) {
		sprintf(r, "%.3gMB", (n/(double)MEGABYTE));
	}
	else if(n>=KILOBYTE) {
		sprintf(r, "%.3gKB", (n/(double)KILOBYTE));
	}
	else
	{
		sprintf(r, "%" PRIuMAX " bytes", n);
	}

// DEBUG:	
//printf("DEBUG: _to_human %" PRIuMAX " bytes=%s\n", n, r);
	return r;
}



uintmax_t _recursive_dir_size(char* dirname)
/* recursively compute the disk usage of a directory
   returns no bytes used
*/
{
	if(dirname==NULL || !*dirname) return 0;
	
	uintmax_t total=0;

	DIR* mydir=opendir(dirname);
	struct dirent* f=NULL;
	struct stat buf;
	char pathname[FILENAME_MAX]={0};
	while( ( f = readdir( mydir ) ) != NULL ) {
		if(strcmp(f->d_name, ".")==0) continue;
		if(strcmp(f->d_name, "..")==0) continue;
		
		// build up full pathname for stat
			*pathname='\0';
			strcat(pathname, dirname);
			strcat(pathname, "/");
			strcat(pathname, f->d_name);
		
		if(stat(pathname, &buf)==0) {	
			// TODO: exclude hardlinked dirs?
			
			if( S_ISDIR( buf.st_mode ) ) {
				total+=_recursive_dir_size(pathname); // recursive call
			}
			else { // it is a regular file
				total+=buf.st_size;
			}
		} // end if
		else
			printf("%s: error reading object size: %s (file skipped)\n", "ss", pathname);
	} // end while
	closedir(mydir);

	return total;
}
/* ALTERNATIVE; NO ftw() IN MINGW
static unsigned int _recursive_dir_size_total = 0;
int _recursive_dir_size(const char *fpath, const struct stat *sb, int typeflag) {
    _recursive_dir_size_total += sb->st_size;
    return 0;
}
*/


/*
TODO: dynamic allocation
char* filenames[argc];
uintmax_t filesizes[argc];
int N=0; // files count
*/

uintmax_t _recursive_add_files(char* dirname)
{
	// TODO
}



int main( int argc, char* argv[] )
{	
	const char* USAGE =
"ss (build date: " __DATE__ ")\n"
"\n"
"usage: ss [options] <list of objects to select>\n"
/*
#if defined(WINDOWS) || defined(WIN32) || defined(_WIN32)
"  NOTE: on Windows file globbing may not work (for example if the pathname is quoted).\n"
#endif*/
"\n"
"options:\n"
"  -s bin-size   \n"
"  -p precision  \n"
"  -e \"exclude wildcard\" \n"
//"  -exec [command]    action to perform on selected files\n"
//"  -printf [format] -> using format syntax http://man.cx/find
"sub directories handling:\n"
"  -n   don't recurse into directories. Treat subdirs as single files (default)\n"
//"  -r   recurse into directories (NOT YET IMPLEMENTED)\n"
"  -i   skip directories\n"
"  -v   verbose\n"
"\n"
"  sizes can be specified using human suffixes and string constants.\n"
"  Supported strings are: fd, cd-r, dvd-r (default), dvd+r, dvd-rdl, and dvd+rdl\n"
"  If not specified precision is adapted automatically according to the current bin size.\n"
"\n"
"examples:\n"
"  ss -v -s 2MB * \n"
"  ss -v -s dvd+r -e \"*.bak\" /home/user/files_to_burn/* \n"
"\n";
	
	// input datas
		char* filenames[argc];
		uintmax_t filesizes[argc]; memset(filesizes, 0, sizeof(filesizes));
		
		int N=0; // files count
		uintmax_t W=DVD_MINUS_R_SIZE; // capacity in bytes. default=DVD-R size
		uintmax_t P=MEGABYTE; // precision in bytes. default=1MB
	
	// parse args
    // TODO? use getopt from unistd.h|Argtable-1.4 http://sourceforge.net/projects/argtable/
		if( ( argc < 2 ) ||
			( strcmp(argv[1], "-h") == 0) ||
			( strcmp(argv[1], "--help") == 0 )
		  )
		{
			puts(USAGE);
			exit(0);
		}

		int a;
		char dirs_handling_mode='n'; // it can be: n=don't recurse, r=recurse, i=ignore/skip
		int verbose=0; // default off
		char* exclude_wildcard = NULL;
		for(a = 1; a < argc; a++)
		{
			if(strcmp(argv[a], "-v")==0)
			{
				verbose=1;
			}
			else if(strcmp(argv[a], "-s")==0)
			{
				// size option detected, parse what follows...
				a++;
				W = _to_bytes(argv[a]);
				if((int)W<=0) {
					printf("%s fatal error: invalid bin size specified: %s\n", argv[0], argv[a]);
					exit(1);
				}
			}
			else if(strcmp(argv[a], "-p")==0)
			{
				// precision option detected, parse what follows...
				a++;
				P = _to_bytes(argv[a]);
				if(P<=0) {
					printf("%s fatal error: invalid precision specified: %s\n", argv[0], argv[a]);
					//P=MEGABYTE; // restore default
					exit(1);
				}
			}
			else if(strcmp(argv[a], "-n")==0)
			{
				dirs_handling_mode='n';
			}
			else if(strcmp(argv[a], "-r")==0)
			{
				dirs_handling_mode='r';
			}
			else if(strcmp(argv[a], "-i")==0)
			{
				dirs_handling_mode='i';
			}
			else if(strcmp(argv[a], "-e")==0)
			{
				a++;
				exclude_wildcard = argv[a];
			}
			else // current arg is a file
			{
				char* filename = argv[a];
				
				if( (exclude_wildcard!=NULL) && !(fnmatch(exclude_wildcard, filename, APR_FNM_PATHNAME)) ) { // check exclude wildcard match
					if(verbose) {
						printf("%s: filename exclude wildcard match: %s (file skipped)\n", argv[0], filename);
					}
					continue;
				}

				struct stat buf = {0};
				if( stat(filename, &buf) != 0 ) {
					printf("%s: error reading file size: %s (file skipped)\n", argv[0], filename);
					continue;
				}
				// else stat is successful
				
				if( S_ISDIR( buf.st_mode ) ) {
				
					// TODO: handle sym/hardlinked to avoid infinite recursion
					
					switch(dirs_handling_mode) {
					
						case('i'):
							continue; // break and skip to the next object
						
						case('r'):
						{
							puts("RECURSE OPTION NOT YET IMPLEMENTED, FALLING BACK TO DEFAULT...");
						}
						
						default:	
						{
							filenames[N]=filename;								
							/*
							_recursive_dir_size_total = 0;
							ftw(filename, &_recursive_dir_size, 1); // recursively compute space
							filesizes[N]=_recursive_dir_size_total; 
							*/
							filesizes[N]=_recursive_dir_size(filename);
							N++;
						}
					} // END SWITCH
					
				} else { // it is a regular file
					filenames[N]=filename;
					filesizes[N]=buf.st_size;
					N++;
				} // ENDIF

// DEBUG:
//printf("object read: %s=%" PRIuMAX "\n", filenames[N], filesizes[N]);

			}
		} // END FOR
		
		if(verbose) {
			printf("%s: %d objects read\n", argv[0], N);
			printf("%s: bin size is: %s\n", argv[0], _to_human(W));
		}
		if(N<=1) {
			printf("%s fatal error: not enought files found.\n", argv[0]);
			exit(1);
		}
	
	// reduce precision
		if(W<=10*P)
		{
			if(W>=10*GIGABYTE) P=GIGABYTE;
			else if(W>=10*MEGABYTE) P=MEGABYTE;
			else if(W>=10*KILOBYTE) P=KILOBYTE;
			else P=1;
		}
		if(P!=1)
		{
			W/=P;
			int i;
			for(i=0;i<N;i++)
			{
				if(filesizes[i]/P < 1)
					filesizes[i]=1;
				else
					filesizes[i]/=P;
			}
		}
		if(verbose) {
			printf("%s: precision is: %s\n", argv[0], _to_human(P));
		
	// compute the solution
			printf("%s: processing, please wait... ", argv[0]);
		}
		//int M[N][W]; memset(M, 0, sizeof(M)); // 2FIX: allocazione inefficiente
		uintmax_t i, w;
		// Dynamically Allocating Multidimensional Arrays -> http://c-faq.com/~scs/cclass/int/sx9b.html
		uintmax_t** M = calloc(N, sizeof(uintmax_t*));
		if(M==NULL) {
			printf("%s fatal error: insufficient memory.\n", argv[0]);
			exit(1);
		}
		for(i=0;i<N;i++)
		{
			M[i] = calloc(W, sizeof(uintmax_t));
			if(M[i]==NULL) {
				printf("%s fatal error: insufficient memory.\n", argv[0]);
				exit(1);
			}
		}
		for(i=1;i<N;i++)
			for(w=1;w<W;w++)
			{
				if(filesizes[i]>w)
					M[i][w] = M[i-1][w];
				else
					M[i][w] = max( M[i-1][w], (filesizes[i]+M[i-1][w-filesizes[i]]) );
			}
		if(verbose) {
			puts("done");
	// print the solution
			puts("selected items:");
		}
		i=N-1;
		w=W-1;
		while(i>0)
		{
			if( M[i][w]==M[i-1][w] )
				i--; // discarded
			else
			{
				// item selected
				printf(" %s (~%s)\n", filenames[i], _to_human(filesizes[i]*P));
				w-=filesizes[i];
				i--;
			}
		}
		if(verbose) {
			printf("total size: ~%s\n", _to_human(M[N-1][W-1]*P));
			printf("wasted space: ~%s\n", _to_human((W-M[N-1][W-1])*P));
		}
		
	return 0;
}
